/////////////////////////////////////////////////////////////////////////////
///
/// @file main.cpp
///
/// VM authentication command line.
///
/// @author shrike
///
/// Copyright (c) 2005-2017, Parallels International GmbH
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Parallels International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#include <errno.h>
#include <iostream>
#include <shadow.h>
#include <guestfs.h>
#include "Registry.h"
#include "wincrypt.h"

#include <openssl/des.h>
#include <openssl/md4.h>
#include <openssl/md5.h>
#include <openssl/rc4.h>
#include <openssl/evp.h>
#include <openssl/aes.h>
#include <openssl/err.h>

#include <QSharedPointer>
#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/variant.hpp>
#include <boost/function.hpp>
#include <boost/program_options.hpp>
#include <prlcommon/Logging/Logging.h>
#include <prlcommon/PrlCommonUtilsBase/SysError.h>

namespace po = boost::program_options;

namespace
{
void printHex(const char *name_, const QByteArray& array_)
{
	QString s;
	QByteArray b(array_.toHex());

	for (int i = 0; i < b.size(); i++)
		s.append(b.at(i));

	WRITE_TRACE(DBG_FATAL, "%s: %s", name_, qPrintable(s));
}

} // namespace

namespace Windows
{
struct Account;
///////////////////////////////////////////////////////////////////////////////
// struct SAM declaration

struct SAM
{
	explicit SAM(CRegistry* dataSource_): m_dataSource(dataSource_)
	{
	}

	Prl::Expected<QByteArray, int> getF();
	Prl::Expected<Account, int>
		getV(const std::string& username_);

private:
	QSharedPointer<CRegistry> m_dataSource;
};

///////////////////////////////////////////////////////////////////////////////
// struct Engine declaration

struct Engine
{
	Engine(guestfs_h* handle_, const std::string& root_):
		m_handle(handle_), m_root(root_)
	{
	}

	Prl::Expected<SAM, int> getSAM() const;
	Prl::Expected<QByteArray, int> getSyskey() const;

private:
	int open(const char* path_, QScopedPointer<CRegistry>& dst_) const;

	static const char* s_LSAS[];
	static const int s_PERMUTATION[0x10];

	guestfs_h* m_handle;
	const std::string m_root;
};

} // namespace Windows

///////////////////////////////////////////////////////////////////////////////
// struct Linux declaration

struct Linux
{
	explicit Linux(guestfs_h* handle_): m_handle(handle_)
	{
	}

	Prl::Expected<FILE*, int> pullShadow() const;

private:
	guestfs_h* m_handle;
};

///////////////////////////////////////////////////////////////////////////////
// struct Facade

struct Facade
{
	typedef boost::variant<Linux, Windows::Engine> os_type;

	explicit Facade(guestfs_h* handle_): m_handle(handle_)
	{
	}

	Prl::Expected<void, int> mount(const std::string& root_) const;
	Prl::Expected<std::string, int> getOsRoot() const;
	Prl::Expected<os_type, int> getOsType(const std::string& root_) const;
	Prl::Expected<FILE*, int> download(const char* path_) const;

private:
	static void free(char** list_);

	guestfs_h* m_handle;
};

Prl::Expected<FILE*, int> Facade::download(const char* path_) const
{
	char* n = tmpnam(NULL);
	if (NULL == n)
	{
		WRITE_TRACE(DBG_FATAL, "tmpnam failed");
		return -4;
	}
	if (-1 == guestfs_download(m_handle, path_, n))
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_download: %s", guestfs_last_error(m_handle));
		return -1;
	}
	FILE* output = fopen(n, "r");
	if (NULL == output)
	{
		WRITE_TRACE(DBG_FATAL, "fopen(\"%s\"): %d", path_, errno);
		unlink(n);
		return -4;
	}
	unlink(n);
	return output;
}

void Facade::free(char** list_)
{
	if (NULL == list_)
		return;

	for (char** p = list_; *p != NULL; ++p)
	{
		::free(*p);
	}
	::free(list_);
}

Prl::Expected<void, int> Facade::mount(const std::string& root_) const
{
	char** p = guestfs_inspect_get_mountpoints(m_handle, root_.data());
	if (NULL == p)
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_inspect_get_mountpoints: %s", guestfs_last_error(m_handle));
		return -1;
	}
	typedef std::map<std::string, std::string> map_type;
	map_type m;
	for (char** i = p; *i != NULL; i += 2)
	{
		m[*p] = *(p + 1);
	}
	free(p);
	BOOST_FOREACH(map_type::const_reference x, m)
	{
		guestfs_mount_ro(m_handle,
			x.second.data(), x.first.data());
	}
	return Prl::Expected<void, int>();
}

Prl::Expected<std::string, int> Facade::getOsRoot() const
{
	char** r = guestfs_inspect_os(m_handle);
	if (NULL == r)
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_inspect_os: %s", guestfs_last_error(m_handle));
		return -1;
	}
	if (NULL == r[0])
	{
		free(r);
		WRITE_TRACE(DBG_FATAL, "no operating system found");
		return -2;
	}
	std::string output = r[0];
	free(r);
	return output;
}

Prl::Expected<Facade::os_type, int> Facade::getOsType(const std::string& root_) const
{
	char* t = guestfs_inspect_get_type(m_handle, root_.data());
	if (NULL == t)
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_inspect_get_type: %s", guestfs_last_error(m_handle));
		return -1;
	}
	if (0 == strcmp(t, "linux"))
	{
		::free(t);
		return os_type(Linux(m_handle));
	}
	if (0 == strcmp(t, "windows"))
	{
		::free(t);
		t = guestfs_inspect_get_windows_systemroot(m_handle, root_.data());
		if (NULL == t)
		{
			WRITE_TRACE(DBG_FATAL, "guestfs_inspect_get_windows_systemroot: %s",
				guestfs_last_error(m_handle));
			return -1;
		}
		std::string r = t;
		::free(t);
		return os_type(Windows::Engine(m_handle, r));
	}
	::free(t);
	WRITE_TRACE(DBG_FATAL, "unsupported operating system type");
	return -3;
}

namespace Windows
{
///////////////////////////////////////////////////////////////////////////////
// struct User

struct User
{
	explicit User(const QByteArray& store_): m_store(store_),
		m_info(reinterpret_cast<const USER_V* >(m_store.data()))
	{
	}

	const QString getName() const;
	quint32 getPasswordSize() const
	{
		return m_info->ntpw_size;
	}
	const uint8_t* getPasswordData() const;
	quint8 const (& getPasswordHash() const)[0x10];

private:
	const QByteArray m_store;
	const USER_V* m_info;
};

const QString User::getName() const
{
	return QString(reinterpret_cast<const QChar *>
		(m_info->data + m_info->username_shift),
		m_info->username_size >> 1);
}

quint8 const (& User::getPasswordHash() const)[0x10]
{
	const PSWD_CRYPTED_DATA* d = reinterpret_cast<const PSWD_CRYPTED_DATA* >
		(getPasswordData());

	return d->Hash;
}

const uint8_t* User::getPasswordData() const
{
	return m_info->data + m_info->ntpw_shift;
}

///////////////////////////////////////////////////////////////////////////////
// struct Account

struct Account: User
{
	Account(quint32 rid_, const User& user_): User(user_), m_rid(rid_)
	{
	}

	quint32 getRid() const
	{
		return m_rid;
	}

private:
	const quint32 m_rid;
};

///////////////////////////////////////////////////////////////////////////////
// struct AES

struct AES
{
	AES(const uint8_t* data_, int length_, const uint8_t* salt_):
		m_data(data_), m_length(length_), m_salt(salt_)
	{
	}

	QByteArray operator()(const char* key_) const
	{
		return decrypt(key_);
	}

private:
	void handleError(EVP_CIPHER_CTX *ctx_) const;
	QByteArray decrypt(const char* key_) const;

private:
	const uint8_t* m_data;
	uint32_t m_length;
	const uint8_t* m_salt;
};

void AES::handleError(EVP_CIPHER_CTX *ctx_) const
{
	unsigned long e;

	while ((e = ERR_get_error()))
	{
		char *err = ERR_error_string(e, NULL);
		WRITE_TRACE(DBG_FATAL, "%s\n", err);
	}
	if (ctx_)
		EVP_CIPHER_CTX_free(ctx_);
}

QByteArray AES::decrypt(const char* key_) const
{
	EVP_CIPHER_CTX *ctx;
	int len;

	ctx = EVP_CIPHER_CTX_new();
	if (ctx == NULL)
	{
		handleError(ctx);
		return QByteArray();
	}

	if (!EVP_DecryptInit_ex(ctx, EVP_aes_128_cbc(), NULL,
		reinterpret_cast<const uint8_t* >(key_), m_salt))
	{
		handleError(ctx);
		return QByteArray();
	}

	EVP_CIPHER_CTX_set_key_length(ctx, EVP_MAX_KEY_LENGTH);

	QByteArray output(m_length, 0);
	uint8_t* plaintext = reinterpret_cast<uint8_t* >(output.data());
	if (EVP_DecryptUpdate(ctx, plaintext, &len, m_data, m_length) != 1)
	{
		handleError(ctx);
		return QByteArray();
	}

	int total = len;

	if (EVP_DecryptFinal_ex(ctx, plaintext + len, &len) < 0)
	{
		handleError(ctx);
		return QByteArray();
	}

	total += len;
	output.resize(total);
	EVP_CIPHER_CTX_free(ctx);

	return output;
}

///////////////////////////////////////////////////////////////////////////////
// struct Magic

struct Magic
{
	QByteArray hash(const QString& password_);
	QByteArray hash(quint8 const (&bootkey_)[0x30], const QByteArray& syskey_);
	QByteArray decode(const QByteArray& bootkey_, const Account& for_);
	QByteArray decrypt(const QByteArray& hash_, const Account& for_);
	Prl::Expected<QByteArray, int> getSamKey(QByteArray& domAccF_,
			QByteArray& syskey_, int pwLen_);
private:
	static const_DES_cblock* generate(quint8 const (&src_)[7], DES_cblock& dst_);

	static const char s_ANUM[];
	static const char s_AQWERTY[];
	static const char s_ANTPASSWORD[];
};

const char Magic::s_ANUM[] = "0123456789012345678901234567890123456789";
const char Magic::s_AQWERTY[] = "!@#$%^&*()qwertyUIOPAzxcvbnmQQQQQQQQQQQQ)(*@&%";
const char Magic::s_ANTPASSWORD[] = "NTPASSWORD";

const_DES_cblock* Magic::generate(quint8 const (&src_)[7], DES_cblock& dst_)
{
	dst_[0] = src_[0];
	dst_[1] = (((src_[0] & 0x01) << 6) | (src_[1] >> 2)) << 1;
	dst_[2] = (((src_[1] & 0x03) << 5) | (src_[2] >> 3)) << 1;
	dst_[3] = (((src_[2] & 0x07) << 4) | (src_[3] >> 4)) << 1;
	dst_[4] = (((src_[3] & 0x0F) << 3) | (src_[4] >> 5)) << 1;
	dst_[5] = (((src_[4] & 0x1F) << 2) | (src_[5] >> 6)) << 1;
	dst_[6] = (((src_[5] & 0x3F) << 1) | (src_[6] >> 7)) << 1;
	dst_[7] = (src_[6] & 0x7F) << 1;
	DES_set_odd_parity(&dst_);

	return &dst_;
}

QByteArray Magic::hash(quint8 const (&bootkey_)[0x30], const QByteArray& syskey_)
{
	MD5_CTX md5c;
	RC4_KEY rc4k;
	quint8 md5hash[0x10];

	MD5_Init(&md5c);
	MD5_Update(&md5c, bootkey_, 0x10);
	MD5_Update(&md5c, s_AQWERTY, sizeof(s_AQWERTY));
	MD5_Update(&md5c, syskey_.data(), syskey_.size());
	MD5_Update(&md5c, s_ANUM, sizeof(s_ANUM));
	MD5_Final(md5hash, &md5c);

	QByteArray output(0x20, 0);
	RC4_set_key(&rc4k, sizeof(md5hash), md5hash);
	RC4(&rc4k, output.size(), &bootkey_[0x10],
		reinterpret_cast<quint8* >(output.data()));

	return output;
}

QByteArray Magic::decode(const QByteArray& key_, const Account& for_)
{
	const SAM_HASH* h = reinterpret_cast<const SAM_HASH* >
			(for_.getPasswordData());

	switch (h->revision)
	{
	case 1:
	{
		MD5_CTX md5c;
		RC4_KEY rc4k;
		quint8 md5hash[0x10];

		MD5_Init(&md5c);
		quint32 r = for_.getRid();
		MD5_Update(&md5c, key_.data(), 0x10);
		MD5_Update(&md5c, reinterpret_cast<const quint8* >(&r), sizeof(r));
		MD5_Update(&md5c, s_ANTPASSWORD, sizeof(s_ANTPASSWORD));
		MD5_Final(md5hash, &md5c);

		QByteArray output(sizeof(for_.getPasswordHash()), 0);
		RC4_set_key(&rc4k, sizeof(md5hash), md5hash);
		RC4(&rc4k, output.size(), &for_.getPasswordHash()[0],
				reinterpret_cast<quint8* >(output.data()));
		return output;
	}
	case 2:
	{
		const SAM_HASH_AES *a = reinterpret_cast<const SAM_HASH_AES *>
				(for_.getPasswordData());
		Windows::AES aes(a->data, for_.getPasswordSize() -
				offsetof(SAM_HASH_AES, data), a->salt);
		return aes(key_.data());
	}
	default:
		WRITE_TRACE(DBG_FATAL, "Unknown SAM hash revision %d",
				h->revision);
		return QByteArray();
	}
}

QByteArray Magic::decrypt(const QByteArray& hash_, const Account& for_)
{
	DES_key_schedule des_key1 = DES_key_schedule(), des_key2 = DES_key_schedule();
	DES_cblock des_cb1 = {}, des_cb2 = {};

	quint32 r = for_.getRid();
	quint8 x[7] = {quint8(r), quint8(r >> 8), quint8(r >> 16), quint8(r >> 24)};
	x[4] = x[0];
	x[5] = x[1];
	x[6] = x[2];
	DES_set_key(generate(x, des_cb1), &des_key1);

	quint8 y[7] = {quint8(r >> 24), quint8(r), quint8(r >> 8), quint8(r >> 16)};
	y[4] = y[0];
	y[5] = y[1];
	y[6] = y[2];
	DES_set_key(generate(y, des_cb2), &des_key2);

	QByteArray output(hash_.size(), 0);
	DES_cblock (&o)[2] = *reinterpret_cast<DES_cblock (*)[2]>(output.data());
	DES_cblock (&i)[2] = *reinterpret_cast<DES_cblock (*)[2]>
		(const_cast<char* >(hash_.data()));
	DES_ecb_encrypt(&i[0], &o[0], &des_key1, DES_DECRYPT);
	DES_ecb_encrypt(&i[1], &o[1], &des_key2, DES_DECRYPT);

	return output;
}

QByteArray Magic::hash(const QString& password_)
{
	MD4_CTX x;
	MD4_Init(&x);
	MD4_Update(&x, password_.utf16(), password_.size() * sizeof(ushort));
	QByteArray output(0x10, 0);
	MD4_Final(reinterpret_cast<quint8* >(output.data()), &x);

	return output;
}

Prl::Expected<QByteArray, int> Magic::getSamKey(QByteArray& domAccF_,
		QByteArray& key_, int pwLen_)
{
	char *p = domAccF_.data();
	uint16_t rev = *p;

	if (rev == 2)
	{
		if (sizeof(PSWD_CRYPTED_DATA) == pwLen_)
			return hash(*reinterpret_cast<quint8 const (*)[0x30]>(p + 0x70), key_);

		WRITE_TRACE(DBG_FATAL, "password length mismatch (%d bytes)",
				pwLen_);
	}
	else if (rev == 3)
	{
		SAM_KEY_DATA_AES *a = reinterpret_cast<SAM_KEY_DATA_AES* >(p + 0x68);

		if (a->revision == 2)
		{
			Windows::AES aes(a->data, a->datalen, a->salt);
			return aes(key_.data());
		}
		WRITE_TRACE(DBG_FATAL, "Unknown AES key rvision %d",
				a->revision);
	}
	else
		WRITE_TRACE(DBG_FATAL, "Unknown Account/F revision %d", rev);

	return -1;
}

///////////////////////////////////////////////////////////////////////////////
// struct SAM definition

Prl::Expected<QByteArray, int> SAM::getF()
{
	quint32 a;
	if (PRL_FAILED(m_dataSource->OpenKey("SAM\\Domains\\Account", a)))
		return -1;

	quint32 r = 0x1000;
	QByteArray output(r, 0);
	if (PRL_FAILED(m_dataSource->GetValueData(a, "F", 0,
		reinterpret_cast<quint8* >(output.data()), r, r)))
		return -1;

	output.resize(r);
	return output;
}

Prl::Expected<Account, int> SAM::getV(const std::string& username_)
{
	quint32 a;
	if (PRL_FAILED(m_dataSource->OpenKey("SAM\\Domains\\Account\\Users", a)))
		return -1;

	QString k;
	for (quint32 i = 0; PRL_SUCCEEDED(m_dataSource->EnumKey(a, i, &k, NULL)); ++i)
	{
		if ("Names" == k)
			continue;

		quint32 h;
		if (PRL_FAILED(m_dataSource->OpenKey(k, h, a)))
			continue;

		quint32 r = 0x1000;
		QByteArray v(r, 0);
		if (PRL_FAILED(m_dataSource->GetValueData(h, "V", 0,
			reinterpret_cast<quint8* >(v.data()), r, r)))
			return -1;

		v.resize(r);
		User u(v);
		QString n = u.getName();
		WRITE_TRACE(DBG_INFO, "user (%d/%d) %s rid=%s",
			n.size(), username_.size(), qPrintable(n), qPrintable(k));
		if (!(username_.size() == static_cast<std::size_t>(n.size()) &&
			n.compare(username_.data(), Qt::CaseInsensitive) == 0))
			continue;
                if (1 != sscanf(qPrintable(k), "%X", &r))
		{
			WRITE_TRACE(DBG_FATAL, "failed to get RID for %s", qPrintable(k));
			return -1;
		}
		return Account(r, u);
	}
	WRITE_TRACE(DBG_FATAL, "Cannot find user's %s account record", username_.data());
	return -5;
}

///////////////////////////////////////////////////////////////////////////////
// struct Engine definition

const char* Engine::s_LSAS[] = {"JD", "Skew1", "GBG", "Data"};
const int Engine::s_PERMUTATION[0x10] =
	{0xb, 0x6, 0x7, 0x1, 0x8, 0xa, 0xe, 0x0, 0x3, 0x5, 0x2, 0xf, 0xd, 0x9, 0xc, 0x4};

int Engine::open(const char* path_, QScopedPointer<CRegistry>& dst_) const
{
	std::string x = m_root + path_;
	char* p = guestfs_case_sensitive_path(m_handle, x.data());
	if (NULL == p)
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_case_sensitive_path: %s", guestfs_last_error(m_handle));
		return -1;
	}
	Prl::Expected<FILE*, int> s = Facade(m_handle).download(p);
	free(p);
	if (s.isFailed())
		return s.error();

	dst_.reset(new CRegistryQtFile(s.value()));
	return 0;
}

Prl::Expected<SAM, int> Engine::getSAM() const
{
	QScopedPointer<CRegistry> h;
	int e = open("/system32/config/sam", h);
	if (0 != e)
		return e;

	return SAM(h.take());
}

Prl::Expected<QByteArray, int> Engine::getSyskey() const
{
	QScopedPointer<CRegistry> h;
	int e = open("/system32/config/system", h);
	if (0 != e)
		return e;

	quint32 k, z;
	if (PRL_FAILED(h->OpenKey("ControlSet001\\Control\\Lsa", k)))
		return -1;

	QByteArray t(0x10, 0);
	for (int i = boost::size(s_LSAS); i-- > 0;)
	{
		if (PRL_FAILED(h->OpenKey(s_LSAS[i], z, k)))
			return -1;

		QString u;
		if (PRL_FAILED(h->QueryInfoKey(z, &u, NULL, NULL, NULL, NULL, NULL, NULL)))
			return -1;

		QByteArray a = u.toUtf8();
		a[8] = 0;
		quint32 (&x)[4] = *reinterpret_cast<quint32(*)[4]>(t.data());
		if (1 != sscanf(a.constData(), "%x", &x[i]))
		{
			WRITE_TRACE(DBG_FATAL, "sscanf %s [%d]\n", a.constData(), errno);
			return -1;
		}
	}
	QByteArray output;
	for (int i = 0; i < t.size(); ++i)
	{
		output.push_back(t.at(s_PERMUTATION[i]));
	}

	return output;
}

} // namespace Windows

///////////////////////////////////////////////////////////////////////////////
// struct Linux definition

Prl::Expected<FILE*, int> Linux::pullShadow() const
{
	return Facade(m_handle).download("/etc/shadow");
}

///////////////////////////////////////////////////////////////////////////////
// struct Visitor

struct Visitor: boost::static_visitor<Prl::Expected<bool, int> >
{
	Visitor(const std::string& username_, const std::string& password_):
		m_username(username_), m_password(password_)
	{
	}

	result_type operator()(const Linux& engine_) const;
	result_type operator()(const Windows::Engine& engine_) const;

private:
	const std::string m_username;
	const std::string m_password;
};

Visitor::result_type Visitor::operator()(const Linux& engine_) const
{
	Prl::Expected<FILE*, int> s = engine_.pullShadow();
	if (s.isFailed())
		return s.error();

	std::string w;
	struct spwd *p = NULL;
	while (true)
	{
		p = fgetspent(s.value());
		if (NULL == p)
			break;
		if (0 == m_username.compare(p->sp_namp))
		{
			w = p->sp_pwdp;
			break;
		}
	}
	fclose(s.value());
	if (w.empty())
		return false;

	char* e = crypt(m_password.data(), w.data());
	if (NULL == e)
	{
		WRITE_TRACE(DBG_FATAL, "cannot crypt");
		return -1;
	}
	return 0 == w.compare(e);
}

Visitor::result_type Visitor::operator()(const Windows::Engine& engine_) const
{
	Prl::Expected<QByteArray, int> sysKey = engine_.getSyskey();
	if (sysKey.isFailed())
		return sysKey.error();

	Prl::Expected<Windows::SAM, int> s = engine_.getSAM();
	if (s.isFailed())
		return s.error();

	Prl::Expected<QByteArray, int> domAccF = s.value().getF();
	if (domAccF.isFailed())
		return domAccF.error();
	
	Prl::Expected<Windows::Account, int> v = s.value().getV(m_username);
	if (v.isFailed())
		return v.error();

	Windows::Magic m;
	QByteArray t = m.hash(m_password.data()), u;
	if (v.value().getPasswordSize() == 0)
		return t == u;

	Prl::Expected<QByteArray, int> samKey = m.getSamKey(domAccF.value(),
			sysKey.value(), v.value().getPasswordSize());
	if (samKey.isFailed())
		return v.error();

	QByteArray hash = m.decode(samKey.value(), v.value());
	u = m.decrypt(hash, v.value());

	printHex("Sys key", sysKey.value());
	printHex("Sam key", samKey.value());
	printHex("Hash", hash);
	printHex("Pw hash", t);
	printHex("NtPw hash", u);;

	return t == u;
}

///////////////////////////////////////////////////////////////////////////////
// struct Abstract

struct Abstract: boost::noncopyable
{
	Abstract();
	~Abstract();

	int operator()();
	int addImage(const std::string& path_);
	void setUsername(const std::string& value_)
	{
		m_username = value_;
	}
	void setPassword(const std::string& value_)
	{
		m_password = value_;
	}

private:
	guestfs_h* m_handle;
	std::string m_username;
	std::string m_password;
};

Abstract::Abstract(): m_handle(guestfs_create())
{
	if (NULL == m_handle)
		WRITE_TRACE(DBG_FATAL, "guestfs_create failed");
}

Abstract::~Abstract()
{
	guestfs_close(m_handle);
}

int Abstract::addImage(const std::string& path_)
{
	if (NULL == m_handle)
		return -1;

	if (-1 == guestfs_add_drive_opts(m_handle, path_.data(), GUESTFS_ADD_DRIVE_OPTS_READONLY, 1, -1))
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_add_drive_opts: %s", guestfs_last_error(m_handle));
		return -1;
	}
	return 0;
}

int Abstract::operator()()
{
	if (-1 == guestfs_launch(m_handle))
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_launch: %s", guestfs_last_error(m_handle));
		return -1;
	}
	Facade f(m_handle);
	Prl::Expected<std::string, int> r = f.getOsRoot();
	if (r.isFailed())
		return r.error();

	Prl::Expected<void, int> m = f.mount(r.value());
	if (m.isFailed())
		return m.error();

	Prl::Expected<Facade::os_type, int> t = f.getOsType(r.value());
	if (t.isFailed())
		return t.error();

	Visitor::result_type x = boost::apply_visitor
		(Visitor(m_username, m_password), t.value());
	if (x.isFailed())
		return x.error();

	return !x.value();
}

///////////////////////////////////////////////////////////////////////////////
// struct Director

struct Director
{
	Director(int argc_, char** argv_);

	int operator()(Abstract& builder_);

private:
	po::command_line_parser m_parser;
	po::options_description m_description;
	po::positional_options_description m_positional;
};

Director::Director(int argc_, char** argv_): m_parser(argc_, argv_), m_description("Allowed options")
{
	m_description.add_options()
		("disk,d", po::value<std::vector<std::string> >(), "path to an image file")
		("verbose,v", "verbose")
		("username,u", po::value<std::string>(), "user name")
		("password,p", po::value<std::string>(), "password");
	m_positional.add("username", 1).add("password", 1);
	m_parser.options(m_description).positional(m_positional);
}

int Director::operator()(Abstract& builder_)
{
	do
	{
		po::variables_map m;
		try
		{
			po::store(m_parser.run(), m);
			po::notify(m);
		}
		catch(std::exception& e_)
		{
			std::cerr << e_.what() << std::endl;
			break;
		}
		if (0 == m.count("disk") || m.count("username") == 0)
			break;

		if (0 < m.count("verbose"))
			SetLogLevel(DBG_DEBUG);

		builder_.setUsername(m["username"].as<std::string>());
		BOOST_FOREACH(const std::string& d, m["disk"].as<std::vector<std::string> >())
		{
			int e = builder_.addImage(d);
			if (0 != e)
				return e;
		}
		if (0 < m.count("password"))
		{
			builder_.setPassword(m["password"].as<std::string>());
			return 0;
		}
		std::string w;
		std::cout << "password: ";
		if (std::getline(std::cin, w))
		{
			builder_.setPassword(w);
			return 0;
		}

	} while(false);

	std::cerr << m_description;
	return -1;
}

int main(int argc_, char** argv_)
{
	SetConsoleLogging(false);
	SetLogFileName(GetDefaultLogFilePath(), "vmauth.log");

	Abstract a;
	Director d(argc_, argv_);
	int e = d(a);
	if (0 != e)
		return e;
	  
	return a();
}

